# Copyright 2021 The QHBM Library Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implementation of general QHBMs in TFQ."""

import inspect
import itertools
import numbers
from typing import Any, Callable, Iterable, List, Union

import cirq
import sympy
import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_quantum as tfq


def build_bit_circuit(qubits, ident):
    """Returns exponentiated X gate on each qubit and the exponent symbols."""
    circuit = cirq.Circuit()
    symbols = []
    for n, q in enumerate(qubits):
        new_bit = sympy.Symbol("_bit_{0}_{1}".format(ident, n))
        circuit += cirq.X(q) ** new_bit
        symbols.append(new_bit)
    return circuit, symbols


@tf.function
def unique_with_counts(input_bitstrings, out_idx=tf.int32):
    """Extract the unique bitstrings in the given bitstring tensor.

    Works by converting each bitstring to a 64 bit integer, then using built-in
    `tf.unique_with_counts` on this 1-D array, then mapping these integers back to
    bitstrings. The inputs and outputs are to be related by the same invariants as
    those of `tf.unique_with_counts`,
    y[idx[i]] = input_bitstrings[i] for i in [0, 1,...,rank(input_bitstrings) - 1]

    TODO(zaqqwerty): the signature and return values are designed to be similar
    to those of tf.unique_with_counts.  This function is needed because
    `tf.unique_with_counts` does not work on 2-D tensors.  When it begins to work
    on 2-D tensors, then this function will be deprecated.

    Args:
      input_bitstrings: 2-D `tf.Tensor` of dtype `int8`.  This tensor is
        interpreted as a list of bitstrings.  Bitstrings are required to be
        64 bits or fewer.
      out_idx: An optional `tf.DType` from: `tf.int32`, `tf.int64`. Defaults to
        `tf.int32`.  Specified type of idx and count outputs.

    Returns:
      y: 2-D `tf.Tensor` of dtype `int8` containing the unique 0-axis entries of
        `input_bitstrings`.
      idx: 1-D `tf.Tensor` of dtype `out_idx` such that `idx[i]` is the index in
        `y` containing the value `input_bitstrings[i]`.
      count: 1-D `tf.Tensor` of dtype `out_idx` such that `count[i]` is the number
        of occurences of `y[i]` in `input_bitstrings`.
    """
    print("retracing: unique_with_counts")
    # Convert bitstrings to integers and uniquify those integers.
    input_shape = tf.shape(input_bitstrings)
    mask = tf.cast(input_bitstrings, dtype=tf.int64)
    base = tf.bitwise.left_shift(
        mask, tf.range(tf.cast(input_shape[1], dtype=tf.int64), dtype=tf.int64)
    )
    ints_equiv = tf.reduce_sum(base, 1)
    _, idx, count = tf.unique_with_counts(ints_equiv, out_idx=out_idx)

    # Convert unique integers to corresponding unique bitstrings.
    y = tf.zeros((tf.shape(count)[0], input_shape[1]), dtype=tf.int8)
    y = tf.tensor_scatter_nd_update(y, tf.expand_dims(idx, axis=1), input_bitstrings)

    return y, idx, count


def upgrade_initial_values(
    initial_values: Union[List[numbers.Real], tf.Tensor, tf.Variable]
) -> tf.Variable:
    """Upgrades the given values to a tf.Variable.

    Args:
      initial_values: Numeric values to upgrade.

    Returns:
      The input values upgraded to a fresh `tf.Variable` of dtype `tf.float32`.
    """
    if isinstance(initial_values, tf.Variable):
        initial_values = initial_values.read_value()
    if isinstance(initial_values, (List, tf.Tensor)):
        initial_values = tf.Variable(
            tf.cast(initial_values, tf.float32), dtype=tf.float32
        )
        if len(tf.shape(initial_values)) != 1:
            raise ValueError("Values for QHBMs must be 1D.")
        return initial_values
    raise TypeError(f"Input needs to be a numeric type, got {type(initial_values)}")


def check_base_function(
    fn: Callable[[tf.Tensor, tf.Tensor], Any]
) -> Callable[[tf.Tensor, tf.Tensor], Any]:
    """Checks that a given function is valid for the base QHBM class.

    The base functions take the `theta` parameters of the classical distribution
    and either a bitstring or a number, all as `tf.Tensor`s.

    Args:
      fn: Function to be checked.

    Returns:
      fn: The input `fn` passed through unchanged.
    """
    sig = inspect.signature(fn)
    if len(sig.parameters) != 2:
        raise TypeError("`fn` must be a two argument Callable.")
    return fn


def upgrade_symbols(
    symbols: Union[Iterable[sympy.Symbol], tf.Tensor],
    values: Union[tf.Tensor, tf.Variable],
) -> tf.Tensor:
    """Upgrades symbols and checks for correct shape.

    For a QHBM, there must be a value associated with each symbol.  So this
    function checks that the shape of `values` is the same as that of `symbols`.

    Args:
      symbols: Iterable of `sympy.Symbol`s to upgrade.
      values: Values corresponding to the symbols.

    Returns:
      `tf.Tensor` containing the string representations of the input `symbols`.
    """
    if isinstance(symbols, Iterable):
        if not all([isinstance(s, sympy.Symbol) for s in symbols]):
            raise TypeError("Each entry of `symbols` must be `sympy.Symbol`.")
        symbols_partial_upgrade = [str(s) for s in symbols]
        if len(set(symbols_partial_upgrade)) != len(symbols):
            raise ValueError("All entries of `symbols` must be unique.")
        symbols_upgrade = tf.constant(symbols_partial_upgrade, dtype=tf.string)
        if tf.shape(symbols_upgrade) != tf.shape(values):
            raise ValueError("There must be a symbol for every value.")
        return symbols_upgrade
    raise TypeError("`symbols` must be an iterable of `sympy.Symbol`s.")


def upgrade_circuit(circuit: cirq.Circuit, symbols: tf.Tensor) -> tf.Tensor:
    """Upgrades a circuit and confirms all symbols are present.

    Args:
      circuit: Circuit to convert to tensor.
      symbols: Tensor of strings which are the symbols in `circuit`.

    Returns:
      Single entry 1D tensor of strings representing the input `circuit`.
    """
    if not isinstance(circuit, cirq.Circuit):
        raise TypeError(f"`circuit` must be a `cirq.Circuit`, got {type(circuit)}")
    if not isinstance(symbols, tf.Tensor):
        raise TypeError("`symbols` must be a `tf.Tensor`")
    if symbols.dtype != tf.string:
        raise TypeError("`symbols` must have dtype `tf.string`")
    if set(tfq.util.get_circuit_symbols(circuit)) != {
        s.decode("utf-8") for s in symbols.numpy()
    }:
        raise ValueError(
            "`circuit` must contain all and only the parameters in `symbols`."
        )
    if not circuit:
        raise ValueError(
            "Empty circuit not allowed. "
            "Instead, use identities on all unused qubits."
        )
    return tfq.convert_to_tensor([circuit])


class QHBM(tf.Module):
    """Class for working with QHBM models in TFQ."""

    def __init__(
        self, initial_thetas, energy, sampler, initial_phis, phis_symbols, u, name
    ):
        """Initializes a QHBM with all the required parameters."""
        if not isinstance(name, str):
            raise TypeError("name must be a string")
        super().__init__(name)
        self.thetas = upgrade_initial_values(initial_thetas)
        self.energy_function = check_base_function(energy)
        self.sampler_function = check_base_function(sampler)
        self.phis = upgrade_initial_values(initial_phis)
        self.phis_symbols = upgrade_symbols(phis_symbols, self.phis)
        self.u = upgrade_circuit(u, self.phis_symbols)
        self.u_dagger = upgrade_circuit(u ** -1, self.phis_symbols)

        self.raw_qubits = sorted(u.all_qubits())
        self.qubits = tf.constant([[q.row, q.col] for q in self.raw_qubits])
        raw_bit_circuit, raw_bit_symbols = build_bit_circuit(self.raw_qubits, name)
        self.bit_symbols = upgrade_symbols(
            raw_bit_symbols, tf.ones([len(self.raw_qubits)])
        )
        self.bit_and_u = tfq.layers.AddCircuit()(raw_bit_circuit, append=u)

        # Simulator backends
        self.tfq_sample_layer = tfq.layers.Sample()

    def copy(self):
        return QHBM(
            self.thetas,
            self.energy_function,
            self.sampler_function,
            self.phis,
            [sympy.Symbol(s.decode("utf-8")) for s in self.phis_symbols.numpy()],
            tfq.from_tensor(self.u)[0],
            self.name,
        )

    @tf.function
    def sample_bitstrings(self, num_samples):
        """Returns unique bitstrings and counts from this QHBM's classical dist.

        Args:
          num_samples: number of bitstrings to sample from the classical probability
              distribution.

        Returns:
          unique_samples:
          counts: 1D `tf.Tensor` of dtype `tf.int32` such that `counts[i]` is the
            number of times `unique_samples[i]` was sampled from `qhbm_sampler`.
        """
        print(f"retracing: sample_bitstrings from {self.name}")
        samples = self.sampler_function(self.thetas, num_samples)
        unique_samples, _, counts = unique_with_counts(samples)
        return unique_samples, counts

    @tf.function
    def sample_state_circuits(self, num_samples):
        """Returns tensor of circuits generating pure state samples from this QHBM.

        Args:
          num_samples: number of pure state samples to draw from this QHBM.

        Returns:
          circuit_samples: 1-D `tf.Tensor` containing the string representations of
              the circuits to generate pures state samples from this QHBM.
          bitstring_samples:
          counts: 1-D `tf.Tensor` of type `tf.int32` such that `counts[i]` says how
              many times circuit `circuit_samples[i]` occurs in the sum used to
              approximate this QHBM.
        """
        print(f"retracing: sample_state_circuits from {self.name}")
        bitstring_samples, counts = self.sample_bitstrings(num_samples)
        circuit_samples = self.state_circuits(bitstring_samples)
        return circuit_samples, bitstring_samples, counts

    @tf.function
    def state_circuits(self, samples):
        print(f"retracing: state_circuits from {self.name}")
        u_model_concrete = tfq.resolve_parameters(
            self.bit_and_u, self.phis_symbols, tf.expand_dims(self.phis, 0)
        )
        tiled_u_model_concrete = tf.tile(u_model_concrete, [tf.shape(samples)[0]])
        circuit_samples = tfq.resolve_parameters(
            tiled_u_model_concrete, self.bit_symbols, samples
        )
        return circuit_samples

    @tf.function
    def sample_unresolved_state_circuits(self, num_samples):
        """Returns tensor of circuits generating pure state samples from this QHBM.

        Args:
          num_samples: number of pure state samples to draw from this QHBM.

        Returns:
          circuit_samples: 1-D `tf.Tensor` containing the string representations of
              the circuits to generate pures state samples from this QHBM.
          bitstring_samples:
          counts: 1-D `tf.Tensor` of type `tf.int32` such that `counts[i]` says how
              many times circuit `circuit_samples[i]` occurs in the sum used to
              approximate this QHBM.
        """
        print(f"retracing: sample_unresolved_state_circuits from {self.name}")
        # Inject the unique bitstrings
        bitstring_samples, counts = self.sample_bitstrings(num_samples)
        tiled_u_model = tf.tile(self.bit_and_u, [tf.shape(counts)[0]])
        circuit_samples = tfq.resolve_parameters(
            tiled_u_model, self.bit_symbols, bitstring_samples
        )
        return circuit_samples, bitstring_samples, counts

    @tf.function
    def sample_pulled_back_bitstrings(self, circuit_samples, counts):
        """Returns samples from the pulled back data distribution.

        The inputs represent the data density matrix. The inverse of this QHBM's
        unitary is appended to create the set of circuits representing the
        pulled back data density matrix. Then, the requested number of bitstrings
        are sampled from each circuit.

        Args:
          circuit_samples: 1-D `tf.Tensor` of type `tf.string` which contains
            circuits serialized by `tfq.convert_to_tensor`. These represent pure
            state samples from the data density matrix. Each entry should be unique.
          counts: 1-D `tf.Tensor` of type `tf.int32`, must be the same size as
            `circuit_samples`. Contains the number of samples to draw from each
            circuit.

        Returns:
          ragged_samples: `tf.RaggedTensor` of DType `tf.int8` structured such
              that `ragged_samples[i]` contains `counts[i]` bitstrings.
        """
        print(f"retracing: sample_pulled_back_bitstrings for {self.name}")
        pulled_back_circuits = tfq.layers.AddCircuit()(
            circuit_samples,
            append=tf.tile(self.u_dagger, [tf.shape(circuit_samples)[0]]),
        )
        raw_samples = self.tfq_sample_layer(
            pulled_back_circuits,
            symbol_names=self.phis_symbols,
            symbol_values=tf.tile(
                tf.expand_dims(self.phis, 0), [tf.shape(counts)[0], 1]
            ),
            repetitions=tf.expand_dims(tf.math.reduce_max(counts), 0),
        )
        num_samples_mask = tf.cast((tf.ragged.range(counts) + 1).to_tensor(), tf.bool)
        ragged_samples = tf.ragged.boolean_mask(raw_samples, num_samples_mask)
        return ragged_samples

    @tf.function
    def energy_and_energy_grad(self, bitstring):
        """Calculates the energy its derivative given a bitstring.

        Args:
          bitstring: `tf.Tensor` of type `tf.int8` which represents the bitstring.

        Returns:
          energy: The energy of this bitstring against the QHBMs energy function.
          grad: The derivative of `energy` with respect to the `thetas` parameters
            of this QHBM.
        """
        print(f"retracing: energy_and_energy_grad for {self.name}")
        new_dup_thetas = tf.identity(self.thetas)  # Need fresh tensor for gradient
        with tf.GradientTape() as tape:
            tape.watch(new_dup_thetas)
            energy = self.energy_function(new_dup_thetas, bitstring)
        grad = tape.gradient(energy, new_dup_thetas)
        return energy, grad

    @tf.function
    def pulled_back_energy_expectation(self, circuit_samples, circuit_counts):
        """Calculates the average energy of bitstrings from the pulled back dist."""
        print(f"retracing: pulled_back_energy_expectation for {self.name}")
        ragged_samples_pb = self.sample_pulled_back_bitstrings(
            circuit_samples, circuit_counts
        )
        # safe when all circuits have the same number of qubits
        all_samples_pb = ragged_samples_pb.values.to_tensor()
        unique_samples_pb, _, counts_pb = unique_with_counts(all_samples_pb)
        counts_pb = tf.cast(counts_pb, tf.float32)
        e_list = tf.map_fn(
            lambda x: self.energy_function(self.thetas, x),
            unique_samples_pb,
            fn_output_signature=tf.float32,
        )
        e_list_full = tf.multiply(counts_pb, e_list)
        average_energy = tf.divide(tf.reduce_sum(e_list_full), tf.reduce_sum(counts_pb))
        return average_energy


class ExactQHBM(QHBM):
    """Used for QHBMs with exact access to the classical and quantum models."""

    def __init__(
        self, initial_thetas, energy, sampler, initial_phis, phis_symbols, u, name
    ):
        """Initializes an analytic QHBM."""
        super().__init__(
            initial_thetas, energy, sampler, initial_phis, phis_symbols, u, name
        )
        self.all_strings = tf.constant(
            list(itertools.product([0, 1], repeat=len(self.raw_qubits))), dtype=tf.int8
        )
        self.tfq_unitary_layer = tfq.layers.Unitary()
        self.tfq_expectation_layer = tfq.layers.Expectation()

    def copy(self):
        return ExactQHBM(
            self.thetas,
            self.energy_function,
            self.sampler_function,
            self.phis,
            [sympy.Symbol(s.decode("utf-8")) for s in self.phis_symbols.numpy()],
            tfq.from_tensor(self.u)[0],
            self.name,
        )

    @tf.function
    def all_energies(self):
        """Returns the energy of every bitstring spanned by this QHBM."""
        print(f"retracing: all_energies for {self.name}")
        # TODO(martantonio): get code to be nearly as fast but with less memory
        # overhead.  tf.map_fn seems to get too CPU fragmented.
        return tf.vectorized_map(
            lambda x: self.energy_function(self.thetas, x), self.all_strings
        )

    @tf.function
    def log_partition_function(self):
        """Calculates the logarithm of the partition function of this QHBM.

        Returns:
          Scalar `tf.Tensor` of DType `tf.float32` which is the logarithm of the
            partition function of the EBM.
        """
        print(f"retracing: log_partition_function for {self.name}")
        negative_energies = -1.0 * self.all_energies()
        return tf.reduce_logsumexp(negative_energies)

    @tf.function
    def entropy_function(self):
        """Calculates the entropy of this QHBM.

        Returns:
          Scalar `tf.Tensor` of DType `tf.float32` which is the entropy of the EBM.
        """
        print(f"retracing: entropy_function for {self.name}")
        negative_energies = -1.0 * self.all_energies()
        return tfp.distributions.Categorical(logits=negative_energies).entropy()

    @tf.function
    def eigvals(self):
        """Returns the eigenvalues of the density matrix this QHBM represents.

        The eigenvalues are the probabilities of the EBM, which are the exponentials
        of the energies divided by the partition function.
        """
        print(f"retracing: eigvals for {self.name}")
        negative_energies = -1.0 * self.all_energies()
        return tf.math.softmax(negative_energies)

    @tf.function
    def unitary_matrix(self):
        """Returns diagonalizing unitary of the density matrix this QHBM represents.

        From the wikipedia page on diagonalization,
        https://en.wikipedia.org/wiki/Diagonalizable_matrix :
        A is diagonalizable if there exists P such that P^{-1} A P is diagonal.
        P is then the unitary diagonalizing A.

        In the case of QHBMs, we have A = UDU^{-1}, where D is the matrix whose
        diagonal is the list of probabilities, and U is the quantum circuit.
        Thus U is a valid choice for P, since then: P^{-1} UDU^{-1} P = D.
        """
        print(f"retracing: unitary_matrix for {self.name}")
        return self.tfq_unitary_layer(
            self.u,
            symbol_names=self.phis_symbols,
            symbol_values=tf.expand_dims(self.phis, 0),
        ).to_tensor()[0]

    @tf.function
    def eigvecs(self):
        """Returns the list of eigenvectors of this QHBM."""
        return tf.transpose(self.unitary_matrix())

    @tf.function
    def density_matrix(self):
        """Returns the exact density matrix represented by this QHBM."""
        print(f"retracing: density_matrix for {self.name}")
        probabilities = tf.cast(self.eigvals(), tf.complex64)
        unitary = self.unitary_matrix()
        u_mul_probs = tf.multiply(
            unitary,
            tf.tile(tf.expand_dims(probabilities, axis=0), (tf.shape(unitary)[0], 1)),
        )
        return tf.linalg.matmul(u_mul_probs, tf.linalg.adjoint(unitary))

    @tf.function
    def fidelity(self, sigma: tf.Tensor):
        """Calculate the fidelity between a QHBM and a density matrix.

        Args:
          sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
            matrix in the fidelity calculation.

        Returns:
          A scalar `tf.Tensor` which is the fidelity between the density matrix
            represented by this QHBM and `sigma`.
        """
        print(f"retracing: fidelity for {self.name}")
        e_rho = tf.cast(self.eigvals(), tf.complex128)
        v_rho = tf.cast(self.unitary_matrix(), tf.complex128)
        sqrt_e_rho = tf.sqrt(e_rho)
        v_rho_sqrt_e_rho = tf.multiply(
            v_rho, tf.tile(tf.expand_dims(sqrt_e_rho, 0), (tf.shape(v_rho)[0], 1))
        )
        rho_sqrt = tf.linalg.matmul(v_rho_sqrt_e_rho, tf.linalg.adjoint(v_rho))
        omega = tf.linalg.matmul(
            tf.linalg.matmul(rho_sqrt, tf.cast(sigma, tf.complex128)), rho_sqrt
        )
        # TODO(zaqqwerty): find convincing proof that omega is hermitian,
        # in order to go back to eigvalsh.
        e_omega = tf.linalg.eigvals(omega)
        return tf.cast(
            tf.math.abs(tf.math.reduce_sum(tf.math.sqrt(e_omega))) ** 2, tf.float32
        )


# TODO(#16)
# class QHBMQ(tf.Module):
#     def __init__(
#         self,
#         p,
#         initial_etas,
#         initial_classical_thetas,
#         eta_theta_symbols,
#         energy,
#         sampler,
#         log_partition,
#         entropy,
#         initial_phis,
#         phis_symbols,
#         u,
#         name,
#     ):
#         """Initialize a QHBM with all the required parameters."""
#         if not isinstance(name, str):
#             raise TypeError("name must be a string")
#         super().__init__(name)

#         self.p = tf.constant(p, dtype=tf.int32)

#         if isinstance(initial_etas, tf.Variable):
#             self.etas = tf.Variable(
#                 tf.cast(initial_etas.read_value(), tf.float32), dtype=tf.float32
#             )
#         elif isinstance(initial_etas, (list, np.ndarray, tf.Tensor)):
#           self.etas = tf.Variable(tf.cast(initial_etas, tf.float32), dtype=tf.float32)
#         else:
#             raise TypeError(
#                 "initial_etas needs to be a numeric type, got {}".format(
#                     type(initial_etas)
#                 )
#             )
#         if self.etas.shape[0] != self.p:
#             raise ValueError("must have as many eta values as trotter steps.")

#         if isinstance(initial_classical_thetas, tf.Variable):
#             self.classical_thetas = tf.Variable(
#                 tf.cast(initial_classical_thetas.read_value(), tf.float32),
#                 dtype=tf.float32,
#             )
#         elif isinstance(initial_classical_thetas, (list, np.ndarray, tf.Tensor)):
#             self.classical_thetas = tf.Variable(
#                 tf.cast(initial_classical_thetas, tf.float32), dtype=tf.float32
#             )
#         else:
#             raise TypeError(
#                 "initial_classical_thetas needs to be a numeric type, got {}".format(
#                     type(initial_classical_thetas)
#                 )
#             )

#         if not isinstance(eta_theta_symbols, (list, tuple)) or not all(
#            [isinstance(s, sympy.Symbol) for row_s in eta_theta_symbols for s in row_s]
#         ):
#             raise TypeError(
#                 "eta_theta_symbols must be a 2-D list or tuple of sympy.Symbols."
#             )
#         self.raw_eta_theta_symbols = copy.deepcopy(eta_theta_symbols)
#         self.eta_theta_symbols = tf.constant(
#             [str(s) for row_s in eta_theta_symbols for s in row_s], dtype=tf.string
#         )
#         if (
#             self.classical_thetas.shape[0] * self.etas.shape[0]
#             != self.eta_theta_symbols.shape[0]
#         ):
#             raise ValueError(
#                 "Due to TFQ limitations, "
#                 "there must be as many eta-theta symbols as etas times thetas."
#             )

#         if not isinstance(energy, collections.Callable):
#             raise TypeError("energy must be a function, got {}".format(type(energy)))
#         self.energy_function = energy
#         if not isinstance(sampler, collections.Callable):
#            raise TypeError("sampler must be a function, got {}".format(type(sampler)))
#         self.sampler_function = sampler
#         if not isinstance(log_partition, collections.Callable):
#             raise TypeError(
#                 "log_partition must be a function, got {}".format(type(log_partition))
#             )
#         self.log_partition_function = log_partition
#         if not isinstance(entropy, collections.Callable):
#            raise TypeError("entropy must be a function, got {}".format(type(entropy)))
#         self.entropy_function = entropy
#         if isinstance(initial_phis, tf.Variable):
#             self.phis = tf.Variable(initial_phis.read_value(), dtype=tf.float32)
#         elif isinstance(initial_phis, (list, np.ndarray, tf.Tensor)):
#             self.phis = tf.Variable(initial_phis, dtype=tf.float32)
#         else:
#             raise TypeError(
#                 "initial_phis needs to be a numeric type, got {}".format(
#                     type(initial_phis)
#                 )
#             )
#         if not isinstance(phis_symbols, (list, tuple)) or not all(
#             [isinstance(s, sympy.Symbol) for s in phis_symbols]
#         ):
#             raise TypeError("phis_symbols must be a list or tuple of sympy.Symbols.")
#         self.raw_phis_symbols = copy.deepcopy(phis_symbols)
#       self.phis_symbols = tf.constant([str(s) for s in phis_symbols], dtype=tf.string)
#         if self.phis.shape != self.phis_symbols.shape:
#             raise ValueError("There must be one symbol per quantum circuit parameter")
#         if not isinstance(u, cirq.Circuit):
#             raise TypeError("u must be a cirq.Circuit, got {}".format(type(u)))
#         self.raw_u = copy.deepcopy(u)
#         if set(tfq.util.get_circuit_symbols(u)) != set(
#             [str(s) for s in self.raw_phis_symbols]
#             + [str(s) for row_s in self.raw_eta_theta_symbols for s in row_s]
#         ):
#             raise ValueError(
#                 "u must contain all and only the symbols "
#                 "in phis_symbols and eta_theta_symbols."
#             )
#         self.u = tfq.convert_to_tensor([u])
#         self.u_dagger = tfq.convert_to_tensor([u ** -1])

#         self.raw_qubits = sorted(u.all_qubits())
#       raw_bit_circuit, raw_bit_symbols = util.build_bit_circuit(self.raw_qubits, name)
#         self.qubits = tf.constant([[q.row, q.col] for q in self.raw_qubits])
#         self.num_qubits = tf.constant(len(self.raw_qubits))
#         self.bit_symbols = tf.constant(
#             [str(s) for s in raw_bit_symbols], dtype=tf.string
#         )
#         self.bit_injector = tfq.convert_to_tensor([raw_bit_circuit])

#     def copy(self):
#         return QHBM(
#             self.p,
#             self.etas,
#             self.classical_thetas,
#             self.eta_theta_symbols,
#             self.energy_function,
#             self.sampler_function,
#             self.log_partition_function,
#             self.entropy_function,
#             self.phis,
#             self.raw_phis_symbols,
#             self.raw_u,
#             self.name,
#         )


# @tf.function
# def get_eta_theta_vals(etas, classical_thetas):
#     """Get the quantum circuit params induced by the classical model."""
#     print("retracing: get_eta_theta_vals")
#     p = tf.shape(etas)[0]
#     num_classical_thetas = tf.shape(classical_thetas)[0]
#     classical_thetas_tiled = tf.tile(tf.expand_dims(classical_thetas, 0), [p, 1])
#     etas_tiled = tf.tile(tf.expand_dims(etas, 1), [1, num_classical_thetas])
#     return tf.reshape(etas_tiled * classical_thetas_tiled, [p * num_classical_thetas])


# @tf.function
# def sample_bitstrings(this_qhbm, num_samples):
#     """Returns unique bitstrings and counts from this QHBM's classical dist.

#     Args:
#       this_qhbm: QHBM representation of a density matrix.
#       num_samples: number of bitstrings to sample from the classical probability
#           distribution parameterizing this QHBM.

#     Returns:
#       unique_samples:
#       counts:
#     """
#     print("retracing: sample_bitstrings")
#     samples = this_qhbm.sampler_function(this_qhbm.classical_thetas, num_samples)
#     unique_samples, _, counts = util.unique_with_counts(samples)
#     return unique_samples, counts


# @tf.function
# def sample_state_circuits(this_qhbm, num_samples):
#     """Returns tensor of circuits generating pure state samples from this QHBM.

#     Args:
#       this_qhbm: QHBM representation of a density matrix.
#       num_samples: number of pure state samples to draw from this QHBM.

#     Returns:
#       circuit_samples: 1-D `tf.Tensor` containing the string representations of
#           the circuits to generate pures state samples from this QHBM.
#       bitstring_samples:
#       counts: 1-D `tf.Tensor` of type `tf.int32` such that `counts[i]` says how
#           many times circuit `circuit_samples[i]` occurs in the sum used to
#           approximate this QHBM.
#     """
#     print("retracing: sample_state_circuits")
#     # Fill in quantum circuit values
#     eta_theta_vals = get_eta_theta_vals(this_qhbm.etas, this_qhbm.classical_thetas)
#     total_vals = tf.concat([eta_theta_vals, this_qhbm.phis], 0)
#    total_symbols = tf.concat([this_qhbm.eta_theta_symbols, this_qhbm.phis_symbols], 0)
#     u_model_concrete = tfq.resolve_parameters(
#         this_qhbm.u, total_symbols, tf.expand_dims(total_vals, 0)
#     )

#     # Inject the unique bitstrings
#     bitstring_samples, counts = sample_bitstrings(this_qhbm, num_samples)
#     tiled_bit_injector = tf.tile(this_qhbm.bit_injector, [tf.shape(counts)[0]])
#     bit_injector_samples = tfq.resolve_parameters(
#         tiled_bit_injector, this_qhbm.bit_symbols, bitstring_samples
#     )
#     circuit_samples = tfq.append_circuit(
#         bit_injector_samples, tf.tile(u_model_concrete, [tf.shape(counts)[0]])
#     )
#     return circuit_samples, bitstring_samples, counts


# @tf.function
# def sample_pulled_back_bitstrings(this_qhbm, circuit_samples, counts):
#     """Returns samples from the pulled back data distribution.

#     The inputs represent the data density matrix. The inverse of this QHBM's
#     unitary is appended to create the set of circuits representing the
#     pulled back data density matrix. Then, the requested number of bitstrings
#     are sampled from each circuit.

#     Args:
#       this_qhbm: QHBM representation of a density matrix.
#     """
#     print("retracing: sample_pulled_back_bitstrings")
#     # Fill in quantum circuit values
#     eta_theta_vals = get_eta_theta_vals(this_qhbm.etas, this_qhbm.classical_thetas)
#     total_vals = tf.concat([eta_theta_vals, this_qhbm.phis], 0)
#    total_symbols = tf.concat([this_qhbm.eta_theta_symbols, this_qhbm.phis_symbols], 0)
#     u_dagger_concrete = tfq.resolve_parameters(
#         this_qhbm.u_dagger, total_symbols, tf.expand_dims(total_vals, 0)
#     )
#     pulled_back_circuits = tfq.append_circuit(
#         circuit_samples, tf.tile(u_dagger_concrete, [tf.shape(circuit_samples)[0]])
#     )
#     raw_samples = tfq.layers.Sample()(
#        pulled_back_circuits, repetitions=tf.expand_dims(tf.math.reduce_max(counts), 0)
#     )
#     num_samples_mask = tf.cast((tf.ragged.range(counts) + 1).to_tensor(), tf.bool)
#     ragged_samples = tf.ragged.boolean_mask(raw_samples, num_samples_mask)
#     return ragged_samples


# @tf.function
# def approximate_density_matrix(this_qhbm, num_samples):
#     """Returns an estimate of the density matrix represented by this QHBM.

#     Args:
#       num_samples: number of pure state samples to use when computing the
#           density matrix represented by this QHBM.

#     Returns:
#       density_matrix: 2-D tensor containing a numeric approximation to the
#           density matrix represented by this QHBM.
#     """
#     print("retracing: approximate_density_matrix")
#     state_circuits, _, counts = sample_state_circuits(this_qhbm, num_samples)
#     return util.circuits_and_counts_to_density_matrix(state_circuits, counts)


# @tf.function
# def energy_and_energy_grad(this_qhbm, bitstring):
#     print("retracing: energy_and_energy_grad")
#     # Need fresh tensor for gradient
#     new_dup_thetas = tf.identity(this_qhbm.classical_thetas)
#     with tf.GradientTape() as tape:
#         tape.watch(new_dup_thetas)
#         this_energy = this_qhbm.energy_function(new_dup_thetas, bitstring)
#     this_grad = tape.gradient(this_energy, new_dup_thetas)
#     return this_energy, this_grad


# @tf.function
# def pulled_back_energy_expectation(this_qhbm, circuit_samples, circuit_counts):
#     """Calculates the average energy of bitstrings from the pulled back dist."""
#     print("retracing: pulled_back_energy_expectation")
#     ragged_samples_pb = sample_pulled_back_bitstrings(
#         this_qhbm, circuit_samples, circuit_counts
#     )
#     # safe when all circuits have the same number of qubits
#     all_samples_pb = ragged_samples_pb.values.to_tensor()
#     unique_samples_pb, _, counts_pb = util.unique_with_counts(all_samples_pb)
#     counts_pb = tf.cast(counts_pb, tf.float32)
#     e_list, _ = tf.map_fn(
#         lambda x: energy_and_energy_grad(this_qhbm, x),
#         unique_samples_pb,
#         fn_output_signature=(tf.float32, tf.float32),
#     )
#     e_list_full = tf.multiply(counts_pb, e_list)
#     average_energy = tf.divide(tf.reduce_sum(e_list_full), tf.reduce_sum(counts_pb))
#     return average_energy



# Copyright 2021 The QHBM Library Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Impementations of the QMHL loss and its derivatives."""

import tensorflow as tf
import tensorflow_quantum as tfq

from qhbmlib import qhbm_base


# ============================================================================ #
# Sample based QMHL TODO(zaqqwerty).
# ============================================================================ #


# ============================================================================ #
# Exact QMHL.
# ============================================================================ #


@tf.function
def exact_qmhl_loss(
    qhbm_model: qhbm_base.ExactQHBM,
    target_circuits: tf.Tensor,
    target_counts: tf.Tensor,
):
    """Calculate the QMHL loss of the model against the target.

    Args:
      qhbm_model: Parameterized model density operator.
      target_circuits: 1-D tensor of strings which are serialized circuits.  These
        circuits represent samples from the data density matrix.
      target_counts: 1-D tensor of integers which are the number of samples to
        draw from the data density matrix: `target_counts[i]` is the number of
        samples to draw from `target_circuits[i]`.

    Returns:
      loss: Quantum cross entropy between the target and model.
    """
    print("retracing: qmhl_loss")
    expected = qhbm_model.pulled_back_energy_expectation(target_circuits, target_counts)
    log_partition = qhbm_model.log_partition_function()
    return expected + log_partition


@tf.function
def exact_qmhl_loss_thetas_grad(
    qhbm_model: qhbm_base.ExactQHBM,
    num_model_samples: tf.Tensor,
    target_circuits: tf.Tensor,
    target_counts: tf.Tensor,
):
    """Calculate thetas gradient of the QMHL loss of the model against the target.

    Args:
      qhbm_model: `QHBM` which is the parameterized model density operator.
      num_model_samples: Scalar integer tensor which is the number of bitstrings
        sampled from the classical distribution of `qhbm_model` to average over
        when estimating the model density operator.
      target_circuits: 1-D tensor of strings which are serialized circuits.  These
        circuits represent samples from the data density matrix.
      target_counts: 1-D tensor of integers which are the number of samples to
        draw from the data density matrix: `target_counts[i]` is the number of
        samples to draw from `target_circuits[i]`.

    Returns:
      Stochastic estimate of the gradient of the QMHL loss with respect to the
        classical model parameters.
    """
    print("retracing: exact_qmhl_loss_thetas_grad")

    # Get energy gradients for the classical bitstrings
    unique_samples_c, counts_c = qhbm_model.sample_bitstrings(num_model_samples)
    expanded_counts_c = tf.cast(
        tf.tile(tf.expand_dims(counts_c, 1), [1, tf.shape(qhbm_model.thetas)[0]]),
        tf.dtypes.float32,
    )
    _, e_grad_list_c = tf.map_fn(
        qhbm_model.energy_and_energy_grad,
        unique_samples_c,
        fn_output_signature=(tf.float32, tf.float32),
    )

    # Get energy gradients for the pulled-back data bitstrings
    ragged_samples_pb = qhbm_model.sample_pulled_back_bitstrings(
        target_circuits, target_counts
    )
    # safe when all circuits have the same number of qubits
    all_samples_pb = ragged_samples_pb.values.to_tensor()
    unique_samples_pb, _, counts_pb = qhbm_base.unique_with_counts(all_samples_pb)
    expanded_counts_pb = tf.cast(
        tf.tile(tf.expand_dims(counts_pb, 1), [1, tf.shape(qhbm_model.thetas)[0]]),
        tf.dtypes.float32,
    )
    _, e_grad_list_pb = tf.map_fn(
        qhbm_model.energy_and_energy_grad,
        unique_samples_pb,
        fn_output_signature=(tf.float32, tf.float32),
    )

    # Build theta gradients after reweighting
    e_grad_c_avg = tf.divide(
        tf.reduce_sum(expanded_counts_c * e_grad_list_c, 0),
        tf.cast(tf.reduce_sum(counts_c), tf.float32),
    )
    e_grad_pb_avg = tf.divide(
        tf.reduce_sum(expanded_counts_pb * e_grad_list_pb, 0),
        tf.cast(tf.reduce_sum(counts_pb), tf.float32),
    )
    return tf.math.subtract(e_grad_pb_avg, e_grad_c_avg)


@tf.function
def exact_qmhl_loss_phis_grad(
    qhbm: qhbm_base.ExactQHBM,
    op_tensor: tf.Tensor,
    target_circuits: tf.Tensor,
    target_count: tf.Tensor,
):
    """Calculate phis gradient of the QMHL loss of the model against the target.

    Args:
      qhbm: Parameterized model density operator.
      op_tensor: Result of calling `tfq.convert_to_tensor` on a list of Cirq
        PauliSums of the form `[[op0, op1, ..., opN-1]]`.  Each op should be
        diagonal in the computational basis.  The assumption made is that
        `qhbm.thetas[i]` is the weight of operator `op_tensor[0][i]` in the
        latent modular Hamiltonian.
      target_circuits: 1-D tensor of strings which are serialized circuits.  These
        circuits represent samples from the data density matrix.
      target_counts: 1-D tensor of integers which are the number of samples to
        draw from the data density matrix: `target_counts[i]` is the number of
        samples to draw from `target_circuits[i]`.

    Returns:
      Stochastic estimate of the gradient of the QMHL loss with respect to the
        unitary model parameters.
    """
    print("retracing: exact_qmhl_loss_phis_grad")
    num_circuits = tf.shape(target_circuits)[0]
    circuits_pb = tfq.append_circuit(
        target_circuits, tf.tile(qhbm.u_dagger, [num_circuits])
    )
    new_dup_phis = tf.identity(qhbm.phis)
    tiled_thetas = tf.tile(tf.expand_dims(qhbm.thetas, 0), [num_circuits, 1])
    with tf.GradientTape() as tape:
        tape.watch(new_dup_phis)
        sub_energy_list = tfq.layers.Expectation()(
            circuits_pb,
            symbol_names=qhbm.phis_symbols,
            symbol_values=tf.tile(tf.expand_dims(new_dup_phis, 0), [num_circuits, 1]),
            operators=tf.tile(op_tensor, [num_circuits, 1]),
        )
        # Weight each operator by the corresponding classical model parameter.
        scaled_sub_energy_list = tiled_thetas * sub_energy_list
        # Get the total latent modular Hamiltonian energy for each input circuit.
        pre_e_avg = tf.reduce_sum(scaled_sub_energy_list, 1)
        e_avg = tf.divide(
            tf.reduce_sum(tf.cast(target_count, tf.float32) * pre_e_avg),
            tf.reduce_sum(tf.cast(target_count, tf.float32)),
        )
    return tape.gradient(e_avg, new_dup_phis)


# ============================================================================ #
# QMHL for non-BUDA QHBMs.
# ============================================================================ #


# TODO(#18)
# @tf.function
# def qmhl_loss_thetas_grad(qhbm_model, num_model_samples, target_density):
#     """Calculate thetas gradient of the QMHL loss of the model against the target.

#     Args:
#       qhbm_model: `QHBM` which is the parameterized model density operator.
#       num_model_samples: number of bitstrings sampled from the classical
#         distribution of `qhbm_model` to average over when estimating the model
#         density operator.
#       target_density: `DensityOperator` which is the distribution whose logarithm
#         we are trying to learn.

#     Returns:
#       Stochastic estimate of the gradient of the QMHL loss with respect to the
#         classical model parameters.
#     """
#     print("retracing: qmhl_loss_thetas_grad")

#     # Get energy gradients for the classical bitstrings
#     unique_samples_c, counts_c = qhbm.sample_bitstrings(
#         qhbm_model.sampler_function, qhbm_model.thetas, num_model_samples
#     )
#     expanded_counts_c = tf.cast(
#         tf.tile(tf.expand_dims(counts_c, 1), [1, tf.shape(qhbm_model.thetas)[0]]),
#         tf.dtypes.float32,
#     )

#     def loop(x):
#         return qhbm.energy_and_energy_grad(
#             qhbm_model.energy_function, qhbm_model.thetas, x
#         )

#     _, e_grad_list_c = tf.map_fn(
#         loop, unique_samples_c, fn_output_signature=(tf.float32, tf.float32)
#     )

#     # Get energy gradients for the pulled-back data bitstrings
#     ragged_samples_pb = qhbm.sample_pulled_back_bitstrings(
#         qhbm_model.u_dagger,
#         qhbm_model.phis_symbols,
#         qhbm_model.phis,
#         target_density.circuits,
#         target_density.counts,
#     )
#     # safe when all circuits have the same number of qubits
#     all_samples_pb = ragged_samples_pb.values.to_tensor()
#     unique_samples_pb, _, counts_pb = qhbm_base.unique_with_counts(all_samples_pb)
#     expanded_counts_pb = tf.cast(
#         tf.tile(tf.expand_dims(counts_pb, 1), [1, tf.shape(qhbm_model.thetas)[0]]),
#         tf.dtypes.float32,
#     )
#     _, e_grad_list_pb = tf.map_fn(
#         loop, unique_samples_pb, fn_output_signature=(tf.float32, tf.float32)
#     )

#     # Build theta gradients after reweighting
#     e_grad_c_avg = tf.divide(
#         tf.reduce_sum(expanded_counts_c * e_grad_list_c, 0),
#         tf.cast(tf.reduce_sum(counts_c), tf.float32),
#     )
#     e_grad_pb_avg = tf.divide(
#         tf.reduce_sum(expanded_counts_pb * e_grad_list_pb, 0),
#         tf.cast(tf.reduce_sum(counts_pb), tf.float32),
#     )
#     return tf.math.subtract(e_grad_pb_avg, e_grad_c_avg)


# @tf.function
# def phis_grad_sub_func(
#     i,
#     qhbm_energy_function,
#     qhbm_thetas,
#     qhbm_u_dagger,
#     qhbm_phis_symbols,
#     qhbm_phis,
#     circuit_samples,
#     circuit_counts,
#     num_phis,
#     eps,
# ):
#     """phis_grad_sub_func."""
#     print("retracing: phis_grad_sub_func")
#     p_axis = tf.one_hot(i, num_phis, dtype=tf.float32)
#     perturbation = p_axis * eps
#     forward = qhbm.pulled_back_energy_expectation(
#         qhbm_energy_function,
#         qhbm_thetas,
#         qhbm_u_dagger,
#         qhbm_phis_symbols,
#         qhbm_phis + perturbation,
#         circuit_samples,
#         circuit_counts,
#     )
#     backward = qhbm.pulled_back_energy_expectation(
#         qhbm_energy_function,
#         qhbm_thetas,
#         qhbm_u_dagger,
#         qhbm_phis_symbols,
#         qhbm_phis - perturbation,
#         circuit_samples,
#         circuit_counts,
#     )
#     return tf.divide(forward - backward, (2.0 * eps))


# @tf.function
# def qmhl_loss_phis_grad(qhbm_model, target_density, eps=0.1):
#     """Calculate phis gradient of the QMHL loss of the model against the target.

#     Args:
#       qhbm_model: `QHBM` which is the parameterized model density operator.
#       target_density: `DensityOperator` which is the distribution whose logarithm
#         we are trying to learn.
#       eps: the size of the finite difference step.

#     Returns:
#       Stochastic estimate of the gradient of the QMHL loss with respect to the
#         unitary model parameters.
#     """
#     print("retracing: qmhl_loss_phis_grad")
#     num_phis = tf.shape(qhbm_model.phis)[0]

#     def loop(x):
#         phis_grad_sub_func(
#             x,
#             qhbm_model.energy_function,
#             qhbm_model.thetas,
#             qhbm_model.u_dagger,
#             qhbm_model.phis_symbols,
#             qhbm_model.phis,
#             target_density.circuits,
#             target_density.counts,
#             num_phis,
#             eps,
#         )

#     return tf.map_fn(loop, tf.range(num_phis), fn_output_signature=tf.float32)



# Copyright 2021 The QHBM Library Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Impementations of the VQT loss and its derivatives."""

import tensorflow as tf
import tensorflow_quantum as tfq

from qhbmlib import qhbm_base


# ============================================================================ #
# Sample-based VQT.
# ============================================================================ #


@tf.function
def vqt_loss(
    qhbm: qhbm_base.QHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Estimates the VQT loss function between a QHBM and a Hamiltonian.

    The two step process follows the VQT loss estimator described in the paper.
    First, samples are drawn from the classical probability distribution of the
    input QHBM.  Then, two quantities are estimated from these samples: the
    entropy of the distribution is calculated by a frequency estimator, and the
    energy is calculated via the Monte-Carlo estimator defined in the paper.

    Args:
      qhbm: The QHBM against which the loss is to be calculated.
      num_samples: The number of samples to draw from the classical probability
        distribution of the QHBM.
      beta: Inverse temperature of the target thermal state.
      hamiltonian: Hamiltonian of the target thermal state.
        It is a 2-D tensor of strings, the result of calling
        `tfq.convert_to_tensor` on a list of lists of cirq.PauliSum which has only
        one term, `[[op]]`.

    Returns:
      Estimate of the VQT loss between the input QHBM and target thermal state.
    """
    samples, counts = qhbm.sample_bitstrings(num_samples)
    prob_terms = tf.cast(counts, tf.float32) / tf.cast(num_samples, tf.float32)
    log_prob_terms = tf.math.log(prob_terms)
    minus_entropy_approx = tf.reduce_sum(prob_terms * log_prob_terms)
    num_circuits = tf.shape(counts)[0]
    state_circuits = qhbm.state_circuits(samples)
    tiled_hamiltonian = tf.tile(hamiltonian, [num_circuits, 1])
    expectations = tf.reshape(
        tfq.layers.SampledExpectation()(
            state_circuits,
            symbol_names=tf.constant([], dtype=tf.string),
            symbol_values=tf.tile(
                tf.constant([[]], dtype=tf.float32), [num_circuits, 1]
            ),
            operators=tiled_hamiltonian,
            repetitions=tf.expand_dims(counts, 1),
        ),
        [num_circuits],
    )
    energy_approx = tf.reduce_sum(expectations * prob_terms)
    return minus_entropy_approx + beta * energy_approx


@tf.function
def vqt_loss_thetas_grad(
    qhbm: qhbm_base.QHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Sample-based estimate of classical param gradient of VQT loss."""
    print("retracing: vqt_loss_thetas_grad")
    # Sample circuits to average the gradient over.
    circuits, bitstrings, counts = qhbm.sample_state_circuits(num_samples)
    num_samples = tf.cast(num_samples, tf.float32)
    num_circuits = tf.shape(circuits)[0]
    counts = tf.cast(counts, tf.float32)

    # Compute the per bitstring difference between beta H and energy
    tiled_hamiltonian = tf.tile(hamiltonian, [num_circuits, 1])
    expectations = tf.reshape(
        tfq.layers.SampledExpectation()(
            circuits,
            symbol_names=tf.constant([], dtype=tf.string),
            symbol_values=tf.tile(
                tf.constant([[]], dtype=tf.float32), [num_circuits, 1]
            ),
            operators=tiled_hamiltonian,
            repetitions=tf.expand_dims(counts, 1),
        ),
        [num_circuits],
    )
    beta_h = beta * expectations
    energies, energy_grads = tf.map_fn(
        qhbm.energy_and_energy_grad,
        bitstrings,
        fn_output_signature=(tf.float32, tf.float32),
    )
    diff_list = beta_h - energies

    part_a = tf.divide(tf.reduce_sum(counts * diff_list), num_samples)
    part_b = tf.divide(
        tf.reduce_sum(tf.expand_dims(counts, 1) * energy_grads, 0), num_samples
    )
    prod_list = tf.expand_dims(diff_list, 1) * energy_grads
    part_c = tf.divide(
        tf.reduce_sum(tf.expand_dims(counts, 1) * prod_list, 0), num_samples
    )

    return part_a * part_b - part_c


@tf.function
def vqt_loss_phis_grad(
    qhbm: qhbm_base.QHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Sample-based estimate of quantum circuit param gradient of VQT loss."""
    print("retracing: vqt_loss_phis_grad")
    circuits, _, counts = qhbm.sample_unresolved_state_circuits(num_samples)
    num_circuits = tf.shape(counts)[0]
    new_dup_phis = tf.identity(qhbm.phis)
    with tf.GradientTape() as tape:
        tape.watch(new_dup_phis)
        sub_energy_list = tf.reshape(
            tfq.layers.Expectation()(
                circuits,
                symbol_names=qhbm.phis_symbols,
                symbol_values=tf.tile(
                    tf.expand_dims(new_dup_phis, 0), [num_circuits, 1]
                ),
                operators=tf.tile(hamiltonian, [num_circuits, 1]),
            ),
            [num_circuits],
        )
        e_avg = tf.divide(
            tf.reduce_sum(tf.cast(counts, tf.float32) * sub_energy_list),
            tf.cast(num_samples, tf.float32),
        )
    this_grad = tape.gradient(e_avg, new_dup_phis)
    return beta * this_grad


# ============================================================================ #
# Exact VQT.
# ============================================================================ #


def _tiled_expectation(circuits: tf.Tensor, hamiltonian: tf.Tensor):
    """Calculates the expectation value for every circuit.

    Args:
      circuits: 1-D tensor of strings which are TFQ serialized circuits with
        no free parameters.
      hamiltonian: 2-D tensor of strings, the result of calling
        `tfq.convert_to_tensor` on a list of lists of cirq.PauliSum which has only
        one term, `[[op]]`.  Will be tiled along the 0th dimension, to be measured
        against each entry of `circuits`.

    Returns:
      1-D tensor of floats which are the expectation values of the single op in
        `hamiltonian` measured against each circuit in `circuits`.
    """
    num_circuits = tf.shape(circuits)[0]
    measurements = tf.tile(hamiltonian, [num_circuits, 1])
    return tf.reshape(
        tfq.layers.Expectation()(
            circuits,
            symbol_names=tf.constant([], dtype=tf.string),
            symbol_values=tf.tile(
                tf.constant([[]], dtype=tf.float32), [tf.shape(circuits)[0], 1]
            ),
            operators=measurements,
        ),
        [num_circuits],
    )


@tf.function
def exact_vqt_loss(
    qhbm: qhbm_base.ExactQHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Calculates the VQT loss function between an ExactQHBM and a hamiltonian."""
    print("retracing: exact_vqt_loss")
    circuits, _, counts = qhbm.sample_state_circuits(num_samples)
    sub_energy_list = _tiled_expectation(circuits, hamiltonian)
    weighted_energy_list = tf.cast(counts, tf.float32) * sub_energy_list
    e_avg = tf.divide(
        tf.reduce_sum(weighted_energy_list), tf.cast(num_samples, tf.float32)
    )
    return beta * e_avg - qhbm.entropy_function()


@tf.function
def exact_vqt_loss_thetas_grad(
    qhbm: qhbm_base.ExactQHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Gradient of exact VQT loss function with respect to classical params."""
    print("retracing: exact_vqt_loss_thetas_grad")
    # Sample circuits to average the gradient over.
    circuits, bitstrings, counts = qhbm.sample_state_circuits(num_samples)
    num_samples = tf.cast(num_samples, tf.float32)
    counts = tf.cast(counts, tf.float32)

    # Compute the per bitstring difference between beta H and energy
    beta_h = beta * _tiled_expectation(circuits, hamiltonian)
    energies, energy_grads = tf.map_fn(
        qhbm.energy_and_energy_grad,
        bitstrings,
        fn_output_signature=(tf.float32, tf.float32),
    )
    diff_list = beta_h - energies

    part_a = tf.divide(tf.reduce_sum(counts * diff_list), num_samples)
    part_b = tf.divide(
        tf.reduce_sum(tf.expand_dims(counts, 1) * energy_grads, 0), num_samples
    )
    prod_list = tf.expand_dims(diff_list, 1) * energy_grads
    part_c = tf.divide(
        tf.reduce_sum(tf.expand_dims(counts, 1) * prod_list, 0), num_samples
    )

    return part_a * part_b - part_c


@tf.function
def exact_vqt_loss_phis_grad(
    qhbm: qhbm_base.ExactQHBM,
    num_samples: tf.Tensor,
    beta: tf.Tensor,
    hamiltonian: tf.Tensor,
):
    """Gradient of exact VQT loss function with respect to circuit params."""
    print("retracing: exact_vqt_loss_phis_grad")
    circuits, _, counts = qhbm.sample_unresolved_state_circuits(num_samples)
    num_circuits = tf.shape(counts)[0]
    new_dup_phis = tf.identity(qhbm.phis)
    with tf.GradientTape() as tape:
        tape.watch(new_dup_phis)
        sub_energy_list = tf.reshape(
            tfq.layers.Expectation()(
                circuits,
                symbol_names=qhbm.phis_symbols,
                symbol_values=tf.tile(
                    tf.expand_dims(new_dup_phis, 0), [num_circuits, 1]
                ),
                operators=tf.tile(hamiltonian, [num_circuits, 1]),
            ),
            [num_circuits],
        )
        e_avg = tf.divide(
            tf.reduce_sum(tf.cast(counts, tf.float32) * sub_energy_list),
            tf.cast(num_samples, tf.float32),
        )
    this_grad = tape.gradient(e_avg, new_dup_phis)
    return beta * this_grad


# ============================================================================ #
# Version of VQT that takes a list of QHBMs as the target Hamiltonian.
# ============================================================================ #

# TODO(#14)
# @tf.function
# def get_avg_sub_energy(
#     full_y_samples_j, qhbm_energy_function, qhbm_thetas_j, bit_sub_indices_j
# ):
#     print("retracing: get_avg_sub_energy")
#     sub_samples = tf.gather(full_y_samples_j, bit_sub_indices_j, axis=1)
#     unique_sub_samples, _, counts = qhbm_base.unique_with_counts(sub_samples)
#     counts = tf.cast(counts, tf.float32)
#     unique_sub_energies, _ = tf.map_fn(
#         lambda x: qhbm.energy_and_energy_grad(qhbm_energy_function, qhbm_thetas_j, x),
#         unique_sub_samples,
#         fn_output_signature=(tf.float32, tf.float32),
#     )
#     sub_energies = tf.multiply(counts, unique_sub_energies)
#     return tf.divide(tf.reduce_sum(sub_energies), tf.reduce_sum(counts))


# def build_sub_term_energy_func(qhbm_vqt, qhbm_hamiltonian_list):
#     bit_sub_indices_list = [
#         util.get_bit_sub_indices(qhbm_vqt.qubits, this_qhbm.qubits)
#         for this_qhbm in qhbm_hamiltonian_list
#     ]
#     qhbm_thetas = [this_qhbm.thetas for this_qhbm in qhbm_hamiltonian_list]
#     qhbm_u_dagger = [this_qhbm.u_dagger for this_qhbm in qhbm_hamiltonian_list]
#     qhbm_phis_symbols=[this_qhbm.phis_symbols for this_qhbm in qhbm_hamiltonian_list]
#     qhbm_phis = [this_qhbm.phis for this_qhbm in qhbm_hamiltonian_list]
#     qhbm_energy_list = [
#         this_qhbm.energy_function for this_qhbm in qhbm_hamiltonian_list
#     ]

#     def inner_build_sub_term_energy_func(
#         energy_list, thetas, sub_indices, daggers, phis_s, phis
#     ):
#         list_of_get_avg_energy_lists = []

#         def inner_jth_sampled_circuits_energy_list(
#             j, i_energy_list, i_thetas, i_sub_indices, i_daggers, i_phis_s, i_phis
#         ):
#             @tf.function
#             def jth_sampled_circuits_energy_list(circuits, counts):
#                 print("retracing: jth_sampled_circuits_energy_list_{}".format(j))
#                 return tf.map_fn(
#                     lambda x: get_avg_sub_energy(
#                         x.to_tensor(), energy_list[j], thetas[j], sub_indices[j]
#                     ),
#                     qhbm.sample_pulled_back_bitstrings(
#                         daggers[j], phis_s[j], phis[j], circuits, counts
#                     ),
#                     fn_output_signature=tf.float32,
#                 )

#             return jth_sampled_circuits_energy_list

#         for j, _ in enumerate(energy_list):
#             list_of_get_avg_energy_lists.append(
#                 inner_jth_sampled_circuits_energy_list(
#                     j, energy_list, thetas, sub_indices, daggers, phis_s, phis
#                 )
#             )

#         @tf.function
#         def sub_term_energy_func(
#             vqt_circuits, vqt_circuits_counts, le=list_of_get_avg_energy_lists
#         ):
#             print("retracing: sub_term_energy_func")
#             avg_energy_lists = tf.stack(
#                 [f(vqt_circuits, vqt_circuits_counts) for f in le]
#             )
#             return tf.reduce_sum(avg_energy_lists, 0)

#         return sub_term_energy_func

#     return inner_build_sub_term_energy_func(
#         qhbm_energy_list,
#         qhbm_thetas,
#         bit_sub_indices_list,
#         qhbm_u_dagger,
#         qhbm_phis_symbols,
#         qhbm_phis,
#     )


# @tf.function
# def vqt_qhbm_loss(qhbm_vqt, num_vqt_samples, sub_term_energy_func):
#     print("retracing: vqt_qhbm_loss")
#     vqt_circuits, _, vqt_circuits_counts = qhbm_vqt.sample_state_circuits(
#         num_vqt_samples,
#     )
#     sub_energy_list = sub_term_energy_func(vqt_circuits, vqt_circuits_counts)
#     e_avg = tf.divide(
#         tf.reduce_sum(tf.cast(vqt_circuits_counts, tf.float32) * sub_energy_list),
#         num_vqt_samples,
#     )
#     return e_avg - qhbm_vqt.entropy_function()


# @tf.function
# def vqt_qhbm_loss_thetas_grad(qhbm_vqt, num_vqt_samples, sub_term_energy_func):
#     print("retracing: vqt_qhbm_loss_thetas_grad")
#     # Build components of the gradient
#     (
#         vqt_circuits,
#         vqt_bitstrings,
#         vqt_circuits_counts,
#     ) = qhbm_vqt.sample_state_circuits(
#         num_vqt_samples,
#     )
#     vqt_expanded_circuits_counts = tf.cast(
#         tf.expand_dims(vqt_circuits_counts, 1), tf.dtypes.float32
#     )
#     vqt_energies, vqt_energy_grads = tf.map_fn(
#         lambda x: qhbm_vqt.energy_and_energy_grad(x),
#         vqt_bitstrings,
#         fn_output_signature=(tf.float32, tf.float32),
#     )

#     # Compute the per vqt bitstring diff list
#     sub_energy_list = sub_term_energy_func(vqt_circuits, vqt_circuits_counts)
#     diff_list = sub_energy_list - vqt_energies
#     expanded_diff_list = tf.expand_dims(diff_list, 1)

#     # Compute the positive term in the gradient
#     diff_list_avg = tf.divide(
#         tf.reduce_sum(tf.cast(vqt_circuits_counts, tf.float32) * diff_list),
#         num_vqt_samples,
#     )
#     vqt_energy_grads_avg = tf.divide(
#         tf.reduce_sum(vqt_expanded_circuits_counts * vqt_energy_grads, 0),
#         num_vqt_samples,
#     )
#     positive = vqt_energy_grads_avg * diff_list_avg

#     # Compute the negative term in the gradient
#     prod_list = vqt_energy_grads * expanded_diff_list
#     negative = tf.divide(
#         tf.reduce_sum(vqt_expanded_circuits_counts * prod_list, 0), num_vqt_samples
#     )

#     return positive - negative


# @tf.function
# def phis_grad_sub_func(
#     i,
#     qhbm,
#     num_vqt_samples,
#     D,
#     eps,
#     sub_term_energy_func,
# ):
#     p_axis = tf.one_hot(i, D, dtype=tf.float32)
#     perturbation = p_axis * eps

#     # Forward
#     original_phis = qhbm.phis.read_value()
#     qhbm.phis.assign_add(perturbation)
#     (vqt_circuits_forward,_,vqt_circuits_counts_forward) = qhbm.sample_state_circuits(
#         num_vqt_samples,
#     )
#     sub_energy_list = sub_term_energy_func(
#         vqt_circuits_forward, vqt_circuits_counts_forward
#     )
#     forward = tf.divide(
#         tf.reduce_sum(
#             tf.cast(vqt_circuits_counts_forward, tf.float32) * sub_energy_list
#         ),
#         num_vqt_samples,
#     )

#     # Backward
#     qhbm.phis.assign_add(-2.0 * perturbation)
#     (
#         vqt_circuits_backward,
#         _,
#         vqt_circuits_counts_backward,
#     ) = qhbm.sample_state_circuits(
#         num_vqt_samples,
#     )
#     sub_energy_list = sub_term_energy_func(
#         vqt_circuits_backward, vqt_circuits_counts_backward
#     )
#     backward = tf.divide(
#         tf.reduce_sum(
#             tf.cast(vqt_circuits_counts_backward, tf.float32) * sub_energy_list
#         ),
#         num_vqt_samples,
#     )
#     qhbm.phis.assign(original_phis)

#     return tf.divide(forward - backward, (2.0 * eps))


# @tf.function
# def vqt_qhbm_loss_phis_grad(qhbm_vqt, num_vqt_samples, sub_term_energy_func, eps=0.1):
#     print("retracing: vqt_qhbm_loss_phis_grad")
#     D = tf.shape(qhbm_vqt.phis)[0]
#     return tf.map_fn(
#         lambda x: phis_grad_sub_func(
#             x,
#             qhbm_vqt,
#             num_vqt_samples,
#             D,
#             eps,
#             sub_term_energy_func,
#         ),
#         tf.range(D),
#         fn_output_signature=tf.float32,
#     )



# Copyright 2021 The QHBM Library Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Module for defining and sampling from EBMs."""

import collections
import itertools

import cirq
import scipy
import tensorflow as tf
import tensorflow_probability as tfp


@tf.function
def probability_to_logit(probability):
    print("retracing: probability_to_logit")
    p = tf.cast(probability, tf.dtypes.float32)
    return tf.math.log(p) - tf.math.log(1 - p)


@tf.function
def logit_to_probability(logit_in):
    print("retracing: logit_to_probability")
    logit = tf.cast(logit_in, tf.dtypes.float32)
    return tf.math.divide(tf.math.exp(logit), 1 + tf.math.exp(logit))


def build_bernoulli(num_nodes, identifier):
    @tf.function
    def energy_bernoulli(logits, bitstring):
        """Calculate the energy of a bitstring against a product of Bernoullis.

        Args:
          logits: 1-D tf.Tensor of dtype float32 containing the
            logits for each Bernoulli factor.
          bitstring: 1-D tf.Tensor of dtype bool of the
            form [x_0, ..., x_n-1].  Must be the same shape as thetas.

        Returns:
          energy: 0-D tf.Tensor of dtype float32 containing the
            energy of the bitstring calculated as
            sum_i[ln(1+exp(logits_i)) - x_i*logits_i].
        """
        print("retracing: energy_bernoulli_{}".format(identifier))
        bitstring = tf.cast(bitstring, dtype=tf.float32)
        return tf.reduce_sum(tf.nn.sigmoid_cross_entropy_with_logits(bitstring, logits))

    @tf.function
    def sampler_bernoulli(thetas, num_samples):
        """Sample bitstrings from a product of Bernoullis.

        Args:
          thetas: 1 dimensional `tensor` of dtype `float32` containing the
            logits for each Bernoulli factor.
          bitstring: 1 dimensional `tensor` of dtype `bool` of the form
            [x_0, ..., x_n-1] so that x is a bitstring.

        Returns:
          bitstrings: `tensor` of dtype `bool` and shape [num_samples, bits]
            where bitstrings are sampled according to
            p(bitstring | thetas) ~ exp(-energy(bitstring | thetas))
        """
        print("retracing: sampler_bernoulli_{}".format(identifier))
        return tfp.distributions.Bernoulli(logits=thetas, dtype=tf.dtypes.int8).sample(
            num_samples
        )

    @tf.function
    def log_partition_bernoulli(thetas):
        print("retracing: log_partition_bernoulli_{}".format(identifier))
        # The result is always zero given our definition of the energy.
        return tf.constant(0.0)

    @tf.function
    def entropy_bernoulli(thetas):
        """Calculate the entropy of a product of Bernoullis.

        Args:
            thetas: 1 dimensional `tensor` of dtype `float32` containing the
              logits for each Bernoulli factor.

        Returns:
          entropy: 0 dimensional `tensor` of dtype `float32` containing the
            entropy (in nats) of the distribution.
        """
        print("retracing: entropy_bernoulli_{}".format(identifier))
        return tf.reduce_sum(
            tfp.distributions.Bernoulli(logits=thetas, dtype=tf.dtypes.int8).entropy()
        )

    return (
        energy_bernoulli,
        sampler_bernoulli,
        log_partition_bernoulli,
        entropy_bernoulli,
        num_nodes,
    )


def build_boltzmann(num_nodes, identifier):

    if num_nodes > 30:
        raise ValueError("Analytic Boltzmann sampling methods fail past 30 bits.")

    def get_all_boltzmann_sub(num_nodes, identifier):
        flat_spins_mask = tf.cast(
            tf.reshape(
                tf.linalg.band_part(tf.ones([num_nodes, num_nodes]), 0, -1)
                - tf.linalg.diag(tf.ones(num_nodes)),
                num_nodes * num_nodes,
            ),
            tf.bool,
        )

        @tf.function
        def boltzmann_bits_to_spins(x):
            print("retracing: boltzmann_bits_to_spins_{}".format(identifier))
            return tf.subtract(
                tf.ones(num_nodes, dtype=tf.int8),
                tf.constant(2, dtype=tf.int8) * tf.cast(x, tf.int8),
            )

        @tf.function
        def energy_boltzmann(thetas, x_in):
            print("retracing: energy_boltzmann_{}".format(identifier))
            spins = tf.cast(boltzmann_bits_to_spins(x_in), tf.float32)
            bias_term = tf.reduce_sum(thetas[:num_nodes] * spins)
            w_slice = thetas[num_nodes:]
            spins_outer = tf.matmul(
                tf.transpose(tf.expand_dims(spins, 0)), tf.expand_dims(spins, 0)
            )
            spins_flat = tf.reshape(spins_outer, [num_nodes * num_nodes])
            interaction_spins = tf.boolean_mask(spins_flat, flat_spins_mask)
            interaction_term = tf.reduce_sum(w_slice * interaction_spins)
            return bias_term + interaction_term

        all_strings = tf.constant(
            list(itertools.product([0, 1], repeat=num_nodes)), dtype=tf.int8
        )

        @tf.function
        def all_energies(thetas):
            print("retracing: all_energies_{}".format(identifier))
            return tf.map_fn(
                lambda x: energy_boltzmann(thetas, x),
                all_strings,
                fn_output_signature=tf.float32,
            )

        @tf.function
        def all_exponentials(thetas):
            print("retracing: all_exponentials_{}".format(identifier))
            return tf.math.exp(
                tf.multiply(tf.constant(-1, dtype=tf.float32), all_energies(thetas))
            )

        @tf.function
        def partition_boltzmann(thetas):
            print("retracing: partition_boltzmann_{}".format(identifier))
            return tf.reduce_sum(all_exponentials(thetas))

        @tf.function
        def log_partition_boltzmann(thetas):
            print("retracing: log_partition_boltzmann_{}".format(identifier))
            return tf.math.log(partition_boltzmann(thetas))

        @tf.function
        def all_probabilities(thetas):
            print("retracing: all_probabilities_{}".format(identifier))
            return all_exponentials(thetas) / partition_boltzmann(thetas)

        @tf.function
        def sampler_boltzmann(thetas, num_samples):
            print("retracing: sampler_boltzmann_{}".format(identifier))
            raw_samples = tfp.distributions.Categorical(
                logits=tf.multiply(
                    tf.constant(-1, dtype=tf.float32), all_energies(thetas)
                ),
                dtype=tf.int32,
            ).sample(num_samples)
            return tf.gather(all_strings, raw_samples)

        @tf.function
        def entropy_boltzmann(thetas):
            print("retracing: entropy_boltzmann_{}".format(identifier))
            these_probs = all_probabilities(thetas)
            these_logs = tf.math.log(these_probs)
            return -1.0 * tf.reduce_sum(these_probs * these_logs)

        return (
            energy_boltzmann,
            sampler_boltzmann,
            log_partition_boltzmann,
            entropy_boltzmann,
            ((num_nodes ** 2 - num_nodes) // 2) + num_nodes,
        )

    return get_all_boltzmann_sub(num_nodes, identifier)


# ============================================================================ #
# K-local EBM tools.
# ============================================================================ #


@tf.function
def bits_to_spins(x, n_bits):
    print("retracing: bits_to_spins")
    return tf.subtract(
        tf.ones(n_bits, dtype=tf.int8), tf.constant(2, dtype=tf.int8) * x
    )


def get_parity_index_list(n_bits, k):
    if k < 1:
        raise ValueError("The locality of interactions must be at least 1.")
    if k > n_bits:
        raise ValueError("The locality cannot be greater than the number of bits.")
    index_list = [i for i in range(n_bits)]
    return tf.constant(list(itertools.combinations(index_list, k)))


def get_single_locality_parities(n_bits, k):
    indices = get_parity_index_list(n_bits, k)

    @tf.function
    def single_locality_parities(spins):
        print("retracing: single_locality_parities")
        return tf.math.reduce_prod(tf.gather(spins, indices), axis=1)

    return single_locality_parities


def get_single_locality_operators(qubits, k):
    index_list = get_parity_index_list(len(qubits), k)
    op_list = []
    for indices in index_list:
        this_op = cirq.PauliSum().from_pauli_strings(1.0 * cirq.I(qubits[0]))
        for i in indices:
            this_op *= cirq.Z(qubits[i])
        op_list.append(this_op)
    return op_list


def get_all_operators(qubits, max_k):
    """Operators corresponding to `get_klocal_energy_function`"""
    op_list = []
    for k in range(1, max_k + 1):
        op_list += get_single_locality_operators(qubits, k)
    return op_list


def get_all_parities(n_bits, max_k):
    func_list = []
    for k in range(1, max_k + 1):
        func_list.append(get_single_locality_parities(n_bits, k))

    @tf.function
    def all_parities(spins, func_list=func_list):
        print("retracing: all_parities")
        return tf.concat([f(spins) for f in func_list], axis=0)

    return all_parities


def get_klocal_energy_function_num_values(n_bits, max_k):
    n_vals = 0
    for i in range(1, max_k + 1):
        n_vals += scipy.special.comb(n_bits, i, exact=True)
    return n_vals


def get_klocal_energy_function(n_bits, max_k):
    all_parities = get_all_parities(n_bits, max_k)

    @tf.function
    def klocal_energy_function(thetas, x):
        print("retracing: klocal_energy_function")
        spins = bits_to_spins(x, n_bits)
        parities = all_parities(spins)
        return tf.reduce_sum(tf.math.multiply(thetas, tf.cast(parities, tf.float32)))

    return klocal_energy_function


# ============================================================================ #
# Swish neural network tools.
# ============================================================================ #


def get_swish_net_hidden_width(num_bits):
    return num_bits + 1 + 2


def get_initial_layer(num_bits):
    """Linear initial layer."""
    w_in = num_bits
    w_out = get_swish_net_hidden_width(num_bits)

    @tf.function
    def initial_layer(thetas, x):
        print("retracing: initial_layer")
        mat = tf.reshape(thetas[: w_in * w_out], [w_out, w_in])
        bias = thetas[w_in * w_out : w_in * w_out + w_out]
        return tf.linalg.matvec(mat, x) + bias

    return initial_layer


def get_hidden_layer(num_bits, i):
    """Swish hidden unit."""
    w = get_swish_net_hidden_width(num_bits)

    @tf.function
    def hidden_layer(thetas, x):
        print("retracing: hidden_layer_{}".format(i))
        mat = tf.reshape(thetas[: w ** 2], [w, w])
        bias = thetas[w ** 2 : w ** 2 + w]
        return tf.nn.swish(tf.linalg.matvec(mat, x) + bias)

    return hidden_layer


def get_final_layer(num_bits):
    """Linear final layer."""
    w_in = get_swish_net_hidden_width(num_bits)
    w_out = 1

    @tf.function
    def final_layer(thetas, x):
        print("retracing: final_layer")
        mat = tf.reshape(thetas[: w_in * w_out], [w_out, w_in])
        bias = thetas[w_in * w_out : w_in * w_out + w_out]
        return tf.reduce_sum(tf.linalg.matvec(mat, x) + bias)

    return final_layer


def get_swish_num_values(num_bits, num_layers):
    h_w = get_swish_net_hidden_width(num_bits)
    n_init_params = num_bits * h_w + h_w
    n_hidden_params = h_w ** 2 + h_w
    n_hidden_params_total = n_hidden_params * num_layers
    n_final_params = h_w + 1
    return n_init_params + n_hidden_params_total + n_final_params


def get_swish_network(num_bits, num_layers):
    """Any function mapping [0,1]^n to R^m can be approximated by
    a Swish network with hidden layer width n+m+2."""
    h_w = get_swish_net_hidden_width(num_bits)
    n_init_params = num_bits * h_w + h_w
    n_hidden_params = h_w ** 2 + h_w
    n_hidden_params_total = n_hidden_params * num_layers

    this_initial_layer = get_initial_layer(num_bits)

    def identity(thetas, x):
        return x

    hidden_func = identity

    def get_hidden_stack_inner(previous_func, i):
        def current_hidden_stack(thetas, x):
            cropped_thetas = thetas[i * n_hidden_params : (i + 1) * n_hidden_params]
            return get_hidden_layer(num_bits, i)(
                cropped_thetas, previous_func(thetas, x)
            )

        return current_hidden_stack

    for i in range(num_layers):
        hidden_func = get_hidden_stack_inner(hidden_func, i)

    this_final_layer = get_final_layer(num_bits)

    @tf.function
    def swish_network(thetas, x):
        print("retracing: swish_network")
        x = tf.cast(x, tf.float32)
        return this_final_layer(
            thetas[n_init_params + n_hidden_params_total :],
            hidden_func(
                thetas[n_init_params : n_init_params + n_hidden_params_total],
                this_initial_layer(thetas[:n_init_params], x),
            ),
        )

    return swish_network


# ============================================================================ #
# Tools for analytic sampling from small energy functions.
# ============================================================================ #


def get_ebm_functions(num_bits, energy_function, ident):
    """Gets functions for exact calculations on energy based models over bits.

    Energy based models (EBMs) are defined by a parameterized energy function,
    E_theta(b), which maps bitstrings to real numbers.  This energy function
    corresponds to a probability distribution
    p(b) = exp(-1.0 * E_theta(b)) / sum_b exp(-1.0 * E_theta(b))

    Args:
      num_bits: number of bits in the samples from the ebm.
      energy_function: function accepting a 1-D `tf.Tensor` of floats and
        a 1-D `tf.Tensor` of bools.  The floats are parameters of an energy
        calculation, and the bools are the bitstring whose energy is calculated.
      ident: Python `str` used to identify functions during tracing.

    Returns:
      sampler_function: function for getting samples from the EBM.
      log_partition_function: function to calculate the natural logarithm of the
        partition function of the EBM.
      entropy_function: function for calculating the entropy of the EBM.
    """
    all_strings = tf.constant(
        list(itertools.product([0, 1], repeat=num_bits)), dtype=tf.int8
    )

    @tf.function
    def all_energies(thetas):
        """Given the EBM parameters, returns the energy of every bitstring."""
        print("retracing: all_energies_{}".format(ident))
        # TODO(zaqqwerty): get code to be nearly as fast but with less memory
        # overhead.  tf.map_fn seems to get too CPU fragmented.
        return tf.vectorized_map(lambda x: energy_function(thetas, x), all_strings)

    @tf.function
    def sampler_function(thetas, num_samples):
        """Samples from the EBM.

        Args:
          thetas: `tf.Tensor` of DType `tf.float32` which are the parameters of
            the EBM calculation.
          num_samples: Scalar `tf.Tensor` of DType `tf.int32` which is the number
            of samples to draw from the EBM.

        Returns:
          `tf.Tensor` of DType `tf.int8` of shape [num_samples, num_bits] which is
            a list of samples from the EBM.
        """
        print("retracing: sampler_function_{}".format(ident))
        negative_energies = -1.0 * all_energies(thetas)
        raw_samples = tfp.distributions.Categorical(
            logits=negative_energies, dtype=tf.int32
        ).sample(num_samples)
        return tf.gather(all_strings, raw_samples)

    @tf.function
    def log_partition_function(thetas):
        """Calculates the logarithm of the partition function of the EBM.

        Args:
          thetas: `tf.Tensor` of DType `tf.float32` which are the parameters of
            the EBM calculation.

        Returns:
          Scalar `tf.Tensor` of DType `tf.float32` which is the logarithm of the
            partition function of the EBM.
        """
        print("retracing: log_partition_function_{}".format(ident))
        negative_energies = -1.0 * all_energies(thetas)
        return tf.reduce_logsumexp(negative_energies)

    @tf.function
    def entropy_function(thetas):
        """Calculates the entropy of the EBM.

        Args:
          thetas: `tf.Tensor` of DType `tf.float32` which are the parameters of
            the EBM calculation.

        Returns:
          Scalar `tf.Tensor` of DType `tf.float32` which is the entropy of the EBM.
        """
        print("retracing: entropy_function_{}".format(ident))
        negative_energies = -1.0 * all_energies(thetas)
        return tfp.distributions.Categorical(logits=negative_energies).entropy()

    return sampler_function, log_partition_function, entropy_function


# ============================================================================ #
# Tools for MCMC sampling from arbitrary energy functions.
# ============================================================================ #


BitstringKernelResults = collections.namedtuple(
    "BitstringKernelResults",
    ["target_log_prob", "log_acceptance_correction", "energy_function_params"],
)


class BitstringKernel(tfp.mcmc.TransitionKernel):
    def __init__(self, proposal_function, energy_function):
        super().__init__()
        self.proposal_function = proposal_function
        self.energy_function = energy_function

    @tf.function
    def one_step(self, current_state, previous_kernel_results):
        print("retracing: one_step")
        next_state = self.proposal_function(current_state)
        kernel_results = BitstringKernelResults(
            target_log_prob=-1.0
            * self.energy_function(
                previous_kernel_results.energy_function_params, next_state
            ),
            log_acceptance_correction=tf.constant(0.0),
            energy_function_params=previous_kernel_results.energy_function_params,
        )
        return next_state, kernel_results

    def bootstrap_results(self, init_state, energy_function_params):
        kernel_results = BitstringKernelResults(
            target_log_prob=-1.0
            * self.energy_function(energy_function_params, init_state),
            log_acceptance_correction=tf.constant(0.0),
            energy_function_params=energy_function_params,
        )
        return kernel_results

    def is_calibrated(self):
        return False


MCMCBitstringKernelResults = collections.namedtuple(
    "MCMCBitstringKernelResults",
    [
        "accepted_results",
        "is_accepted",
        "log_accept_ratio",
        "proposed_state",
        "proposed_results",
    ],
)


class MCMCBitstringKernel(tfp.mcmc.TransitionKernel):
    """Adapted from the TFP MetropolisHastings kernel."""

    def __init__(self, inner_kernel):
        super().__init__()
        self.inner_kernel = inner_kernel

    @tf.function
    def one_step(self, current_state, previous_kernel_results):
        print("retracing: one_step")
        [proposed_state, proposed_results] = self.inner_kernel.one_step(
            current_state, previous_kernel_results.accepted_results
        )
        log_accept_ratio = (
            proposed_results.target_log_prob
            - previous_kernel_results.accepted_results.target_log_prob
            + proposed_results.log_acceptance_correction
        )
        # If proposed state reduces likelihood: randomly accept.
        # If proposed state increases likelihood: always accept.
        # I.e., u < min(1, accept_ratio),  where u ~ Uniform[0,1)
        #       ==> log(u) < log_accept_ratio
        log_uniform = tf.math.log(
            tf.random.uniform(
                tf.shape(proposed_results.target_log_prob),
                minval=0,
                maxval=None,
                dtype=proposed_results.target_log_prob.dtype,
            )
        )
        is_accepted = log_uniform < log_accept_ratio
        next_state = tf.where(is_accepted, proposed_state, current_state)
        accepted_results = tf.cond(
            is_accepted,
            lambda: proposed_results,
            lambda: previous_kernel_results.accepted_results,
        )
        kernel_results = MCMCBitstringKernelResults(
            accepted_results=accepted_results,
            is_accepted=is_accepted,
            log_accept_ratio=log_accept_ratio,
            proposed_state=proposed_state,
            proposed_results=proposed_results,
        )
        return next_state, kernel_results

    def bootstrap_results(self, init_state, energy_function_params):
        pkr = self.inner_kernel.bootstrap_results(init_state, energy_function_params)
        x = pkr.target_log_prob
        return MCMCBitstringKernelResults(
            accepted_results=pkr,
            is_accepted=tf.ones_like(x, dtype=tf.bool),
            log_accept_ratio=tf.zeros_like(x),
            proposed_state=init_state,
            proposed_results=pkr,
        )

    def is_calibrated(self):
        return True


def get_bernoulli_proposal_function(flip_prob, num_qubits):
    dist = tfp.distributions.Bernoulli(probs=[flip_prob] * num_qubits, dtype=tf.bool)

    @tf.function
    def bernoulli_proposal_function(current_state):
        print("retracing: bernoulli_proposal_function")
        mask = dist.sample(1)[0]
        return tf.math.logical_xor(current_state, mask)

    return bernoulli_proposal_function


def get_chain(
    proposal_function,
    energy_function,
    num_burnin_steps,
    num_steps_between_results,
    parallel_iterations,
):
    inner_kernel = BitstringKernel(proposal_function, energy_function)
    mcmc_kernel = MCMCBitstringKernel(inner_kernel)

    @tf.function
    def run_chain(last_state, energy_function_params, num_steps):
        print("retracing: run_chain")
        last_state = tf.cast(last_state, tf.bool)
        return tf.cast(
            tfp.mcmc.sample_chain(
                num_results=num_steps,
                num_burnin_steps=num_burnin_steps,
                num_steps_between_results=num_steps_between_results,
                parallel_iterations=parallel_iterations,
                current_state=last_state,
                kernel=mcmc_kernel,
                previous_kernel_results=mcmc_kernel.bootstrap_results(
                    last_state, energy_function_params
                ),
                trace_fn=None,
            ),
            tf.int8,
        )

    return run_chain
